'''
Jesus Mendoza, Colorado School of Mines
First Made: 01-10-19
Last Updated: 06-19-19

This program is intended to control any camera connected to 
a raspberry pi running INDI. This program mainly serves to 
take exposures and cool the camera. It is an adaptation of
the tutorial on taking an exposure and a cooling command supplied
by an INDI user (forum post listed below).

Initial Code Adapted From: https://indilib.org/support/tutorials/166-installing-and-using-the-python-pyndi-client-on-raspberry-pi.html
Cooling Code Adapted From: https://indilib.org/forum/ccds-dslrs/3429-cooling-of-ccd-via-pyindi.html
    
'''
import time
import PyIndi
import threading
import subprocess

'''
The functions listed here are needed because PyIndi calls them
from the base library due to requirements for interacting with 
the INDI server.
'''
class IndiClient(PyIndi.BaseClient):
    def __loader__it__(self):
        super(IndiClient, self).__init__()
    def newDevice(self, d):
        pass
    def newProperty(self, p):
        pass
    def removeProperty(self, p):
        pass
    def newBLOB(self, bp):
        global blobEvent
        print("new BLOB ", bp.name)
        blobEvent.set()
        pass
    def newSwitch(self, svp):
        pass
    def newNumber(self, nvp):
        pass
    def newText(self, tvp):
        pass
    def newLight(self, lvp):
        pass
    def newMessage(self, d, m):
        pass
    def serverConnected(self):
        pass
    def serverDisconnected(self, code):
        pass

def StartIndi(ctype):
    process=subprocess.Popen(['indiserver',ctype])

def StopIndi():
    process.kill()
    
def StartCam():
    # connect the server
    global device_ccd,ccd_connect,indiclient,ccd,temp

    indiclient=IndiClient()
    indiclient.setServer("localhost",7624)

    '''
    The following section connects the indiclient (the interface between
    indiserver and python) to the INDI server. 

    ***Make sure indiserver is running the appropriate driver before the program starts.***

    E.g. In a terminal: indiserver indi_asi_ccd
    '''
    indiclient.connectServer()
    time.sleep(0.1) #Need time to connect to server 1-30
    devices=indiclient.getDevices()[0] # Receive the connected device 1-30
    ccd=devices.getDeviceName() # Gets the name of the device according to the indiserver 1-30

    device_ccd=indiclient.getDevice(ccd) #Retrieves the camera by calling its name from the indiserver
    while not(device_ccd):
        time.sleep(0.5)
        device_ccd=indiclient.getDevice(ccd)
    
    #Connects the camera to the indiclient
    ccd_connect=device_ccd.getSwitch("CONNECTION")
    while not(ccd_connect):
        time.sleep(0.5)
        ccd_connect=device_ccd.getSwitch("CONNECTION")
    if not(device_ccd.isConnected()):
        ccd_connect[0].s=PyIndi.ISS_ON  # the "CONNECT" switch
        ccd_connect[1].s=PyIndi.ISS_OFF # the "DISCONNECT" switch
        indiclient.sendNewSwitch(ccd_connect)
    
    temp = device_ccd.getNumber("CCD_TEMPERATURE")
    while temp==None:
            temp = device_ccd.getNumber("CCD_TEMPERATURE")
            
def CamTemp(cam_temp,tolerance):
    '''
    This section looks for the temperature of the CCD and sends the
    appropriate temperature. Other values can be retrieved and set
    the same way.
    
    https://indilib.org/develop/developer manual/101-standard-properties.html#h2-general
    This link leads to the camera properties that can be retrieved.
    
    The temperature will stay at the desired value until the device has
    been disconnected and the indiclient has been disconnected from
    the server.
    '''
    global temp
    camera_temp=cam_temp
    temp[0].value=camera_temp       # new temperature to reach
    indiclient.sendNewNumber(temp) #self is the indiclient. Sends the desired temperature to the indiserver.
    time.sleep(2.5) #New value is sent but the indiclient needs to time retrieve the current value after sending the desired one.
    h=0
    while temp[0].value>= (camera_temp+tolerance):   #Checks to make sure the temperature is within a tolerance in degrees C (e.g. 1 C) of the set temperature.
        if h % 250==True:
            print(temp[0].value)
            '''
            Prints out the temperature value every 250 checks. (~2.5 seconds)
            '''
        time.sleep(0.01)
        h+=1
    print(temp[0].value)

def CamTempGUI(cam_temp): #Will not wait for camera to reach temp in order to continue
    global temp
    temp[0].value=cam_temp       # new temperature to reach
    indiclient.sendNewNumber(temp) #self is the indiclient. Sends the desired temperature to the indiserver.
    time.sleep(2.5) #New value is sent but the indiclient needs to time retrieve the current value after sending the desired one.

def Temp():
    print(temp[0].value)
    
def CamExposure(exp_times,exp_names):
    global blobEvent, device_ccd, indiclient
    '''
    This section deals with the exposures. The indiclient camera
    interface (device_ccd) first gets the value that is being sent for an
    exposure from the indiserver. It then sets up the blob (exposure data)
    so that it can receive and read the information and receives the blob
    data from the indiserver.
    
    exp_times = exposure times in a list. 
    exp_names = exposure names in a list.
    '''
    ccd_exposure=device_ccd.getNumber("CCD_EXPOSURE")
    while not(ccd_exposure):
        time.sleep(0.5)
        ccd_exposure=device_ccd.getNumber("CCD_EXPOSURE")
     
    # we should inform the indi server that we want to receive the
    # "CCD1" blob from this device
    indiclient.setBLOBMode(PyIndi.B_ALSO, ccd, "CCD1")
    ccd_ccd1=device_ccd.getBLOB("CCD1")
    while not(ccd_ccd1):
        time.sleep(0.5)
        ccd_ccd1=device_ccd.getBLOB("CCD1")
    
    # we use here the threading.Event facility of Python
    # we define an event for newBlob event
    blobEvent=threading.Event()
    blobEvent.clear()

    expnum=0
    while expnum<len(exp_times):
        ccd_exposure[0].value=exp_times[expnum]
        indiclient.sendNewNumber(ccd_exposure) #Sends the exposure time to the indiserver.
        blobEvent.wait() #Waits for the indiserver to send back the exposure.
        
        for blob in ccd_ccd1:
            fits=blob.getblobdata() #Retrieves the image data from the blob as a FITS file format.
            with open(exp_names[expnum]+".fits","wb") as f:
                f.write(fits)
                '''
                Saves the data to a file. The initial info informs the image reader
                that the file is in fits format.
                '''
        blobEvent.clear() #Clears the blob for the next exposure.
        expnum+=1
        print(exp_names[expnum-1]+" has been saved.\n")
    '''
    In order to take successive exposures, there needs to be a
    waiting period. Refer to pyindi tutorial at:
    
    https://indilib.org/develop/indi-python-bindings.html
    
    Copy from            ccd_exposure[expnum].value=exposures[i]
    to                   blobEvent.clear()
    
    to take another image. Change the fits name to not
    overwrite the data of the previous image.
    
    In a while loop, 10 seconds is known to work.
    '''
    
def StopCamera():
    global indiclient, ccd
    indiclient.disconnectDevice(ccd)
    indiclient.disconnectServer()

'''
To use this program:

StartIndi('indi_???_ccd')
StartCam()
CamTemp(#, +-#)
CamExposure([10],['image'])
StopCamera()
StopIndi()

There can be as many exposures as you want (i.e. more
CamExposure() lines or just make lists for the amount
of exposures you want i.e.

5 10s images:
times=[10,10,10,10,10]
names=['image_0' , 'image_1', 'image_2' , 'image_3' , 'image_4']

CamExposure(times,names)
------------------------------------------------------------
For the indi server, you can open a terminal and run the line:

indiserver indi_???_ccd

indefinitely as long as the camera was connected beforehand.
This cam replace the StartIndi() and StopIndi() functions.
If the python script stops before running StopIndi(), the
program will be lost in the background and will not be able
to be closed out of incase you want to open another instance
of an indiserver.

Note: replace ??? with the appropriate letters for the device:

asi for ZWO ccd's
atik for ATIK ccds
sx for Starlight Xpress Cameras

Go to
https://www.indilib.org/devices/ccds.html
for a full list of supported camera types and their repective
indi driver executable.

To view images use astropy.io to get the data and matplotlib
to plot.
'''
